/**
 * Copyright (C) 2003-2004 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

package sync4j.framework.protocol;

import java.util.*;

import sync4j.framework.core.*;
import sync4j.framework.database.Database;
import sync4j.framework.protocol.v11.*;
import sync4j.framework.tools.MD5;

/**
 *
 * @author  Stefano Fornari @ Funambol
 *
 * @version $Id: ProtocolUtil.java,v 1.9 2004/07/07 14:54:12 luigiafassina Exp $
 *
 */
public class ProtocolUtil {

    /** Filters the given commands based on the given class
     *
     * @param commands commands to be filtered
     * @param filterClass selector
     *
     * @return a java.util.List containing the selected commands
     *
     */
    public static ArrayList filterCommands(AbstractCommand[] commands, Class filterClass) {
        ArrayList filteredCommands = new ArrayList();

        for (int i=0; i<commands.length; ++i) {
            if (filterClass.isInstance(commands[i])) {
                filteredCommands.add(commands[i]);
            }
        }

        return filteredCommands;
    }

    /**
     * Filters the given commands based on the given command id. A command is
     * selected if it a ResponseCommand (only ResponseCommand subclasses have
     * got the reference command id) and its command id ref matches <i>cmdId</i>.
     *
     * @param commands commands to be filtered
     * @param cmdId selector
     *
     * @return a java.util.List containing the selected commands
     *
     */
    public static ArrayList filterCommands(AbstractCommand[] commands,
                                           CmdID cmdId               ) {
        ArrayList filteredCommands = new ArrayList();

        for (int i=0; i<commands.length; ++i) {
            if (  (commands[i] instanceof ResponseCommand)
               && ((((ResponseCommand)commands[i]).getCmdID()).equals(cmdId))) {
                filteredCommands.add(commands[i]);
            }
        }

        return filteredCommands;
    }

    /**
     * Filters the given commands based on the given command type and command id.
     * It combines <i>filterCommands(commands, filterClass)</i> and
     * <i>filterCommands(commands, cmdId)</i> returning only the commands
     * that respect both requirements.
     *
     * @param commands commands to be filtered
     * @param filterClass class type selector
     * @param cmdId selector
     *
     * @return a java.util.List containing the selected commands
     *
     */
    public static ArrayList filterCommands(AbstractCommand[] commands   ,
                                           Class             filterClass,
                                           CmdID cmdId                  ) {
        //
        // Since filtering on command identifier seems more selective,
        // filterCommands(commands, cmdId) is called first and than
        // filterCommands(..., filterClass) is called with the returned values.
        //
        ArrayList list = filterCommands(commands, cmdId);
        int size = list.size();
        AbstractCommand [] aCommands = new AbstractCommand[size];
        for (int i=0; i < size; i++) {
            aCommands[i] = (AbstractCommand)list.get(i);
        }
        return filterCommands(
                    /* not compatible with j2me
                    (AbstractCommand[])list.toArray(new AbstractCommand[0]),
                    */
                    aCommands,
                   filterClass
               );
    }

    /**
     * Filters the given Status commands to search the SyncHdr Status.
     *
     * @param commands the Status commands to be filtered
     *
     * @return a java.util.List containing the selected commands
     *
     */
    public static ArrayList filterStatusSyncHdr(Status[] commands) {

        ArrayList filteredCommands = new ArrayList();

        for (int i=0; i<commands.length; ++i) {
            if ( "SyncHdr".equals(commands[i].getCmd()) &&
                 "0".equals(commands[i].getCmdRef())      ) {
                filteredCommands.add(commands[i]);
            }
        }

        return filteredCommands;
    }

    /**
     * Creates and returns and AlertCommand for the synchronization of the
     * given database.
     *
     * @param id the command id - NULL
     * @param noResponse
     * @param credential - NULL
     * @param db the database to be synchronized - NOT NULL
     *
     * @return the AlertCommand object
     */
    public static Alert createAlertCommand(CmdID    id        ,
                                           boolean  noResponse,
                                           Cred     credential,
                                           Database db        ) {
        Item[] items = new Item[1];

        Anchor serverAnchor = db.getServerAnchor();

        Meta meta = new Meta();
        meta.setAnchor(serverAnchor);
        items[0] = new Item(db.getTarget(), 
                            db.getSource(), 
                            meta          , 
                            null          ,  //data
                            false         ); //MoreData

        return new Alert(
                   id            ,
                   noResponse    ,
                   credential    ,
                   db.getMethod(),
                   items
               );
    }

    /**
     * Translates a Target object to a Source object
     *
     * @param target the target object - NULL
     *
     * @return a Source object with the same URI and local name of <i>target</i>
     */
    public static Source target2Source(Target target) {
        if (target == null) return null;

        return new Source(target.getLocURI(), target.getLocName());
    }

    /**
     * Translates a Source object to a Target object
     *
     * @param source the source object - NULL
     *
     * @return a Target object with the same URI and local name of <i>source</i>
     */
    public static Target source2Target(Source source) {
        if (source == null) return null;

        return new Target(source.getLocURI(), source.getLocName());
    }
    
    /** 
     * Extracts the target and source refs from an array of items
     *
     * @param items the items to inspect. If null targetRefs and sourceRefs
     *              remain unchanged
     * @param targetRefs a reference to an array that will contain the references
     *                   to the items' targets
     * @param sourceRefs a reference to an array that will contain the references
     *                   to the items' sources
     *
     */
    public static void extractRefs(Item[] items          , 
                                   TargetRef[] targetRefs,
                                   SourceRef[] sourceRefs) {
        if (items == null) {
            return;
        }
        
        Target t = null;
        Source s = null;
        for (int i=0; i<items.length; ++i) {
            t = items[i].getTarget();
            s = items[i].getSource();
            
            targetRefs[i] = (t != null) ? new TargetRef(t) : null;
            sourceRefs[i] = (s != null) ? new SourceRef(s) : null;
        }
    }
    
    /**
     * Filters a list of commands extracting the ones of the given types.
     *
     * @param commands the list of command to be filtered
     * @param types the command types to extract
     *
     * @return an array of the selected commmands
     */
    public static List filterCommands(List commands, String[] types) {
        StringBuffer sb = new StringBuffer(",");

        for (int i = 0; ((types != null) && (i < types.length)); ++i) {
            sb.append(types[i]).append(',');
        }

        ArrayList selectedCommands = new ArrayList();
        AbstractCommand command = null;
        Iterator i = commands.iterator();
        while (i.hasNext()) {
            command = (AbstractCommand) i.next();

            if (sb.indexOf(',' + command.getName() + ',') >= 0) {
                selectedCommands.add(command);
            }
        }

        return selectedCommands;
    }

    /**
     * Checks if a message require a response from the client.<p>
     * A message requires a response if its body contains commands other than
     * status commands.
     *
     * @param msg the message to check - NOT NULL and properly constructed
     *
     * @return true if the message requires a response, false otherwise
     *
     */
    public static boolean noMoreResponse(SyncML msg) {
        AbstractCommand[] commands =
        (AbstractCommand[])msg.getSyncBody().getCommands().toArray(
        new AbstractCommand[0]);

        for(int i=0; ((commands != null) && (i<commands.length)); ++i) {
            if (!Status.COMMAND_NAME.equals(commands[i].getName())) {
                return false;
            }
        }

        return true;
    }

    /**
     * Generate the next nonce for MD5 authentication
     * The nextNonce for the session has a number format;
     * the nextNonce for the chal element is encoding b64
     *
     * @return NextNonce a new NextNonce object
     */
    public static NextNonce generateNextNonce() {
        return new NextNonce(MD5.getNextNonce());
    }

    /**
     * Returns the header status code of the given message, if specified.<br<
     * The header status code is the first status in the message body. The
     * first message of PCK1 has not any header status code. In this case -1 is
     * returned.
     *
     * @param msg the SyncML message object
     *
     * @return the header status code or -1 if the given message does not
     *         containno any header status command.
     */
    public static int getHeaderStatusCode(SyncML msg) {
        ArrayList cmdList = msg.getSyncBody().getCommands();

        cmdList = filterCommands(
                      (AbstractCommand[])cmdList.toArray(new AbstractCommand[cmdList.size()]),
                      Status.class,
                      new CmdID("1")
                  );

        if (cmdList.size() == 0) {
            return -1;
        }

        return ((Status)cmdList.get(0)).getStatusCode();
    }

    /**
     * Returns the Chal element included in the header status if there is any or
     * null if no chal is given.
     *
     * @param msg the SyncML message object
     *
     * @return Chal element included in the header status if there is any or
     *         null if no chal is given.
     */
    public static Chal getStatusChal(SyncML msg) {
        ArrayList cmdList = msg.getSyncBody().getCommands();

        cmdList = filterCommands(
                      (AbstractCommand[])cmdList.toArray(new AbstractCommand[cmdList.size()]),
                      Status.class,
                      new CmdID("1")
                  );

        if (cmdList.size() == 0) {
            return null;
        }

        return ((Status)cmdList.get(0)).getChal();
    }
    
    /**
     * Sort an array of Status object in according to their cmdRef
     *
     * @param statusToSort an array of Status object
     * 
     * @return an array of sorted Status object
     */
    public static Object[] sortStatusCommand(Object[] statusToSort) {
        StatusComparator comparator = new StatusComparator();
        Arrays.sort(statusToSort, comparator);
        return statusToSort;
    }
    
    /**
     * Sort an array of AbstractCommand object in according to their cmdID
     *
     * @param statusToSort an array of AbstractCommand object
     * 
     * @return an array of sorted AbstractCommand object
     */
    public static Object[] sortAbstractCommand(Object[] cmdToSort) {
        AbstractCmdComparator comparator = new AbstractCmdComparator();
        Arrays.sort(cmdToSort, comparator);
        return cmdToSort;
    }
    
    // --------------------------------------------------------- Private methods

    /**
     * This class compares two Status object in according to their cmdRef
     */
    private static class StatusComparator implements java.util.Comparator {
        private StatusComparator() {
        }

        public int compare(Object o1, Object o2) {
            Object value1 = null;
            Object value2 = null;

            value1 = new Integer(((Status)o1).getCmdRef());
            value2 = new Integer(((Status)o2).getCmdRef());

            return ( (Integer)value1).compareTo((Integer)value2);
        }
    }
    
    /**
     * This class compares two AbstractCommand object in according to their cmdID
     */
    private static class AbstractCmdComparator implements java.util.Comparator {
        private AbstractCmdComparator() {
        }

        public int compare(Object o1, Object o2) {
            Object value1 = null;
            Object value2 = null;

            value1 = new Integer(((AbstractCommand)o1).getCmdID().getCmdID());
            value2 = new Integer(((AbstractCommand)o2).getCmdID().getCmdID());

            return ( (Integer)value1).compareTo((Integer)value2);
        }
    }
}
